use macros::decode_args;
use num_derive::FromPrimitive;
use num_traits::FromPrimitive;
use truth::llir::RawInstr;
use wgpu::naga::FastHashMap;

use crate::game::param::Param;

#[derive(Debug, FromPrimitive, Clone, Copy)]
enum Opcode {
  Nop = 0,

  // control flow
  Delete = 1,
  Ret = 10,
  Call = 11,
  Jump = 12,
  JumpEq = 13,
  JumpNeq = 14,
  CallAsync = 15,
  CallAsyncById = 16,
  KillAsync = 17,
  KillAllAsync = 21,

  // stack math
  StackAlloc = 40,
  PushInt = 42,
  SetInt = 43,
  PushFloat = 44,
  SetFloat = 45,
  AddInt = 50,
  AddFloat = 51,
  SubInt = 52,
  SubFloat = 53,
  MulInt = 54,
  MulFloat = 55,
  DivInt = 56,
  DivFloat = 57,
  ModInt = 58,
  EqualInt = 59,
  EqualFloat = 60,
  NotEqualInt = 61,
  NotEqualFloat = 62,
  LessInt = 63,
  LessFloat = 64,
  LessEqualInt = 65,
  LessEqualFloat = 66,
  GreaterInt = 67,
  GreaterFloat = 68,
  GreaterEqualInt = 69,
  GreaterEqualFloat = 70,
  NotInt = 71,
  NotFloat = 72,
  LogicalOr = 73,
  LogicalAnd = 74,
  BitwiseXor = 75,
  BitwiseOr = 76,
  BitwiseAnd = 77,
  DecrementInt = 78,
  Sin = 79,
  Cos = 80,
  CirclePos = 81,
  ValidRad = 82,
  Wait = 83,
  NegateInt = 84,
  NegateFloat = 85,
  SquareSum = 86,
  GetAngle = 87,
  SquareRoot = 88,
  Linear = 89,
  PointRotate = 90,
  FloatTime = 91,
  Math92 = 92,
  Math93 = 93,

  EnemyCreate = 256,
  EnemyCreateAbsolute = 257,
  EnemySelectAnm = 258,
  EnemySetAnmSprite = 259,
  EnemyCreateMirrored = 260,
  EnemyCreateMirroredAbsolute = 261,
  EnemySetMainAnm = 262,
  EnemyPlayAnm = 263,
  EnemyPlayAnmAbsolute = 264,
  EnemyCreateFiller = 265,
  EnemyCreateFillerAbsolute = 266,
  EnemyCreateFillerMirrored = 267,
  EnemyCreateFillerMirroredAbsolute = 268,
  EnemyPlaySelected = 269,
  EnemySwitchAnm = 275,
  EnemyResetAnm = 276,
  EnemyInst277 = 277,
  EnemyInst278 = 278,

  MovePos = 280,
  MovePosTime = 281,
  MovePosRel = 282,
  MovePosRelTime = 283,
  MoveVel = 284,
  MoveVelTime = 285,
  MoveVelRel = 286,
  MoveVelRelTime = 287,
  MoveCircle = 288,
  MoveCircleTime = 289,
  MoveCircleRel = 290,
  MoveCircleRelTime = 291,
  MoveRand = 292,
  MoveRandRel = 293,
  MoveAdd = 294,
  MoveAddRel = 295,
  MoveEllipse = 300,
  MoveEllipseRel = 301,
  MoveBezier = 305,
  MoveReset = 307,

  SetHurtbox = 320,
  SetHitbox = 321,
  FlagSet = 322,
  FlagClear = 323,
  Movelimit = 324,
  MoveLimitReset = 325,
  ClearExtraDrops = 326,
  AddExtraDrops = 327,
  SetDropArea = 328,
  DropItems = 329,
  SetMainDrop = 330,
  SetHealth = 331,
  SetBoss = 332,
  TimerReset = 333,
  SetInterrupt = 334,
  SetInvulnerable = 335,
  PlaySound = 336,
  ShakeScreen = 337,
  StartDialogue = 338,
  WaitForDialogue = 339,
  WaitForDeath = 340,
  SetTimeout = 341,
  SpellById = 342,
  EndSpell = 343,
  SetChapter = 344,
  KillAllEnemies = 345,
  ProtectPlayer = 346,
  LifeMarker = 347,
  SetByDifficultyInt = 355,
  SetByDifficultyFloat = 356,
  SpellDifficulty = 357,
  SpellDifficultyM1 = 358,
  SpellDifficultyM2 = 359,

  SpawnerReset = 400,
  SpawnerEnable = 401,
  SpawnerSetSprite = 402,
  SpawnerSetOffset = 403,
  SpawnerSetAngle = 404,
  SpawnerSetSpeed = 405,
  SpawnerSetCount = 406,
  SpawnerSetAimMode = 407,
  SpawnerSetSounds,
  SpawnerSet,
  SpawnerSetTransformation,
  SpawnerSetTransformation2,
  SpawnerAddTransformation,
  SpawnerAddTransformation2,
  ClearAllBullets,
  SpawnerCopy,
  ShootLaserAimed,
  CancelBullets,
}

pub enum PrimOp {
  Add,
  Sub,
  Mul,
  Div,
  Equal,
  NotEqual,
  LessThan,
  LessEqual,
  Greater,
  GreaterEqual,
  Not,
  Negate,
}

pub enum Op {
  Nop,
  Delete,
  Return,
  Jump {
    index: usize,
    time: i32,
  },
  JumpEqual {
    index: usize,
    time: i32,
  },
  JumpNotEqual {
    index: usize,
    time: i32,
  },
  CallAsync(String),
  CallAsyncId(String, u32),
  StackAlloc(usize),
  PushInt(Param<i32>),
  SetInt(i32),
  PushFloat(Param<f32>),
  SetFloat(i32),
  IntOp(PrimOp),
  ModuloInt,
  FloatOp(PrimOp),
  Or,
  And,
  BitwiseXor,
  BitwiseOr,
  BitwiseAnd,
  DecrementInt(i32),
  Sin,
  Cos,
  CirclePos {
    x: f32,
    y: f32,
    angle: Param<f32>,
    radius: Param<f32>,
  },

  SetFlag(i32),
  ClearFlag(i32),

  ClearAllBullets,
  CancelBullets(i32),
}

pub struct Instruction {
  time: i32,
  difficulty: u8,
  op: Op,
}

impl Instruction {
  pub fn from_raw(inst: RawInstr, instruction_offsets: &FastHashMap<u32, usize>) -> Self {
    let param_int = |value: i32, index: u16| {
      if inst.param_mask & (1 << index) == 0 {
        Param::Value(value)
      } else if value < -9932 || value > 10000 {
        Param::Value(value)
      } else {
        Param::Variable(value)
      }
    };
    let param_float = |value: f32, index: u16| {
      if inst.param_mask & (1 << index) == 0 {
        Param::Value(value)
      } else if value != value.trunc() && value < -9932. || value > 10000. {
        Param::Value(value)
      } else {
        Param::Variable(value as i32)
      }
    };
    let var_int = |value: i32, index: u16| {
      assert!(inst.param_mask & (1 << index) != 0, "param must be a var");
      assert!(value >= -9932 && value <= 10000, "value must be a var");
      value
    };
    let var_float = |value: f32, index: u16| {
      assert!(inst.param_mask & (1 << index) != 0, "param must be a var");
      assert!(
        value.trunc() == value && value >= -9932. && value <= 10000.,
        "value must be a var"
      );
      value as i32
    };
    let offset_to_index = |value: i32| instruction_offsets[&(value as u32)];

    let args = &inst.args_blob;
    let opcode_raw = inst.opcode;
    let opcode = Opcode::from_u16(opcode_raw).unwrap_or_else(|| todo!("failed to convert opcode {opcode_raw}"));
    let op = match opcode {
      Opcode::Nop => Op::Nop,
      Opcode::Delete => todo!(),
      Opcode::Ret => todo!(),
      Opcode::Call => todo!(),
      Opcode::Jump => {
        let (offset, time) = decode_args!(args, "ot");

        Op::Jump {
          index: offset_to_index(offset),
          time,
        }
      }
      Opcode::JumpEq => {
        let (offset, time) = decode_args!(args, "ot");

        Op::JumpEqual {
          index: offset_to_index(offset),
          time,
        }
      }
      Opcode::JumpNeq => {
        let (offset, time) = decode_args!(args, "ot");

        Op::JumpNotEqual {
          index: offset_to_index(offset),
          time,
        }
      }
      Opcode::CallAsync => todo!(),
      Opcode::CallAsyncById => todo!(),
      Opcode::KillAsync => todo!(),
      Opcode::KillAllAsync => todo!(),
      Opcode::StackAlloc => Op::StackAlloc(decode_args!(args, "S") as usize),
      Opcode::PushInt => todo!(),
      Opcode::SetInt => todo!(),
      Opcode::PushFloat => {
        let float = decode_args!(args, "f");
        Op::PushFloat(param_float(float, 0))
      }
      Opcode::SetFloat => {
        let var = decode_args!(args, "f");
        Op::SetFloat(var_float(var, 0))
      }
      Opcode::FlagSet => {
        let flag = decode_args!(args, "S");
        Op::SetFlag(flag)
      }
      Opcode::FlagClear => {
        let flag = decode_args!(args, "S");
        Op::ClearFlag(flag)
      }
      Opcode::CancelBullets => {
        todo!()
      }
      _ => {
        unimplemented!("opcode {opcode:?} not implemented");
      }
    };

    Self {
      time: inst.time,
      difficulty: inst.difficulty,
      op,
    }
  }
}
